import spikeinterface as si

import os
import time
import numpy as np
from os.path import join
from spiketoolkit.sorters.tools import run_command_and_print_output

def spyking_circus(
        recording,
        output_folder=None,  # Temporary working directory
        probe_file=None,
        file_name=None,
        detect_sign='negative',  # 'negative' - 'positive' - 'both'
        adjacency_radius=100,  # Channel neighborhood adjacency radius corresponding to geom file
        spike_thresh=6,  # Threshold for detection
        template_width_ms=3,  # Spyking circus parameter
        filter=True,
        merge_spikes=True,
        n_cores=None,
        electrode_dimensions=None,
        whitening_max_elts=1000,  # I believe it relates to subsampling and affects compute time
        clustering_max_elts=10000,  # I believe it relates to subsampling and affects compute time
    ):
    try:
        import circus
    except ModuleNotFoundError:
        raise ModuleNotFoundError("\nTo use Spyking-Circus, install spyking-circus: \n\n"
                                  "\npip install spyking-circus"
                                  "\nfor ubuntu install openmpi: "
                                  "\nsudo apt install libopenmpi-dev"
                                  "\nMore information on Spyking-Circus at: "
                                  "\nhttps://spyking-circus.readthedocs.io/en/latest/")
    source_dir = os.path.dirname(os.path.realpath(__file__))

    if output_folder is None:
        output_folder = 'spyking_circus'
    else:
        output_folder = join(output_folder, 'spyking_circus')

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # save prb file:
    if probe_file is None:
        si.saveProbeFile(recording, join(output_folder, 'probe.prb'), format='spyking_circus', radius=adjacency_radius,
                         dimensions=electrode_dimensions)
        probe_file = join(output_folder, 'probe.prb')
    # save binary file
    if file_name is None:
        file_name = 'recording'
    elif file_name.endswith('.npy'):
        file_name = file_name[file_name.find('.npy')]
    np.save(join(output_folder, file_name), recording.getTraces().astype('float32'))

    # set up spykingcircus config file
    with open(join(source_dir, 'config_default.params'), 'r') as f:
        circus_config = f.readlines()
    if merge_spikes:
        auto = 1e-5
    else:
        auto = 0
    circus_config = ''.join(circus_config).format(
        float(recording.getSamplingFrequency()), probe_file, template_width_ms, spike_thresh, detect_sign, filter,
        whitening_max_elts, clustering_max_elts, auto
    )
    with open(join(output_folder, file_name + '.params'), 'w') as f:
        f.writelines(circus_config)

    print('Running spyking circus...')
    t_start_proc = time.time()
    if n_cores is None:
        n_cores = np.maximum(1, int(os.cpu_count()/2))

    cmd = 'spyking-circus {} -c {} '.format(join(output_folder, file_name+'.npy'), n_cores)
    cmd_merge = 'spyking-circus {} -m merging -c {} '.format(join(output_folder, file_name+'.npy'), n_cores)
    print(cmd)
    retcode = run_command_and_print_output(cmd)
    if retcode != 0:
        raise Exception('Spyking circus returned a non-zero exit code')
    print(cmd_merge)
    retcode = run_command_and_print_output(cmd_merge)
    if retcode != 0:
        raise Exception('Spyking circus merging returned a non-zero exit code')
    processing_time = time.time() - t_start_proc
    print('Elapsed time: ', processing_time)
    sorting = si.SpykingCircusSortingExtractor(join(output_folder, file_name))

    return sorting



import spikeinterface as si
import os
from mountainlab_pytools import mdaio
from spiketoolkit.sorters.tools import run_command_and_print_output

def ironclust(*,
    recording, # Recording object
    tmpdir, # Temporary working directory
    detect_sign=-1, # Polarity of the spikes, -1, 0, or 1
    adjacency_radius=-1, # Channel neighborhood adjacency radius corresponding to geom file
    detect_threshold=5, # Threshold for detection
    merge_thresh=.98, # Cluster merging threhold 0..1
    freq_min=300, # Lower frequency limit for band-pass filter
    freq_max=6000, # Upper frequency limit for band-pass filter
    pc_per_chan=3, # Number of pc per channel
    prm_template_name, # Name of the template file
    ironclust_src=None
):      
    if ironclust_src is None:
        ironclust_src=os.getenv('IRONCLUST_SRC',None)
    if not ironclust_src:
        raise Exception('You must either set the IRONCLUST_SRC environment variable, or pass the ironclust_src parameter')
    source_dir=os.path.dirname(os.path.realpath(__file__))

    dataset_dir=tmpdir+'/ironclust_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    si.MdaRecordingExtractor.writeRecording(recording=recording,save_path=dataset_dir)
        
    samplerate=recording.getSamplingFrequency()

    print('Reading timeseries header...')
    HH=mdaio.readmda_header(dataset_dir+'/raw.mda')
    num_channels=HH.dims[0]
    num_timepoints=HH.dims[1]
    duration_minutes=num_timepoints/samplerate/60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(num_channels,num_timepoints,duration_minutes))

    print('Creating .params file...')
    txt=''
    txt+='samplerate={}\n'.format(samplerate)
    txt+='detect_sign={}\n'.format(detect_sign)
    txt+='adjacency_radius={}\n'.format(adjacency_radius)
    txt+='detect_threshold={}\n'.format(detect_threshold)
    txt+='merge_thresh={}\n'.format(merge_thresh)
    txt+='freq_min={}\n'.format(freq_min)
    txt+='freq_max={}\n'.format(freq_max)    
    txt+='pc_per_chan={}\n'.format(pc_per_chan)
    txt+='prm_template_name={}\n'.format(prm_template_name)
    _write_text_file(dataset_dir+'/argfile.txt',txt)
        
    print('Running IronClust...')
    cmd_path = "addpath('{}', '{}/matlab', '{}/mdaio');".format(ironclust_src, ironclust_src, ironclust_src)
    #"p_ironclust('$(tempdir)','$timeseries$','$geom$','$prm$','$firings_true$','$firings_out$','$(argfile)');"
    cmd_call = "p_ironclust('{}', '{}', '{}', '', '', '{}', '{}');"\
        .format(tmpdir, dataset_dir+'/raw.mda', dataset_dir+'/geom.csv', tmpdir+'/firings.mda', dataset_dir+'/argfile.txt')
    cmd='matlab -nosplash -nodisplay -r "{} {} quit;"'.format(cmd_path, cmd_call)
    print(cmd)
    retcode=run_command_and_print_output(cmd)

    if retcode != 0:
        raise Exception('IronClust returned a non-zero exit code')

    # parse output
    result_fname=tmpdir+'/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: '+ result_fname)
    
    firings=mdaio.readmda(result_fname)
    sorting=si.NumpySortingExtractor()
    sorting.setTimesLabels(firings[1,:],firings[2,:])
    return sorting

def _read_text_file(fname):
    with open(fname) as f:
        return f.read()
    
def _write_text_file(fname,str):
    with open(fname,'w') as f:
        f.write(str)


from kbucket import client as kb
from pairio import client as pa
import getpass

def kbucketConfigLocal(write=True):
  pa.setConfig(
      collections=[],
      user='',
      token='',
      read_local=True,write_local=False,read_remote=False,write_remote=False
  )
  kb.setConfig(
      share_ids=[],
      upload_share_id='',
      upload_token='',
      load_local=True,load_remote=False,save_remote=False
  )
  if write:
    pa.setConfig(
        write_local=True,
    )

def kbucketConfigRemote(*,user='spikeforest',share_id='spikeforest.spikeforest2',password=None,write=False):
  pa.setConfig(
      collections=[user],
      user='',
      token='',
      read_local=False,write_local=False,read_remote=True,write_remote=False
  )
  kb.setConfig(
      share_ids=[share_id],
      upload_share_id='',
      upload_token='',
      load_local=True,load_remote=True,save_remote=False
  )
  if write:
    if password is None:
      password=getpass.getpass('Enter the spikeforest password')
    pa.setConfig(
        user=user,
        token=pa.get(collection='spikeforest',key=dict(name='pairio_token',user=user,password=password)),
        write_remote=True
    )
    kb.setConfig(
        upload_share_id=share_id,
        upload_token=pa.get(collection='spikeforest',key=dict(name='kbucket_token',share_id=share_id,password=password)),
        save_remote=True
    )


import spikeinterface as si
import os
from os.path import join
import time
from spiketoolkit.sorters.tools import run_command_and_print_output

def klusta(
        recording, # The recording extractor
        output_folder=None,
        probe_file=None,
        file_name=None,
        threshold_strong_std_factor=5,
        threshold_weak_std_factor=2,
        detect_spikes='negative',
        extract_s_before=16,
        extract_s_after=32,
        n_features_per_channel=3,
        pca_n_waveforms_max=10000,
        num_starting_clusters=50
        ):
    try:
        import klusta
        import klustakwik2
    except ModuleNotFoundError:
        raise ModuleNotFoundError("\nTo use Klusta, install klusta and klustakwik2: \n\n"
                                  "\npip install klusta klustakwik\n"
                                  "\nMore information on klusta at: "
                                  "\nhttps://github.com/kwikteam/phy"
                                  "\nhttps://github.com/kwikteam/klusta")
    source_dir = os.path.dirname(os.path.realpath(__file__))

    if output_folder is None:
        output_folder = 'klusta'
    else:
        output_folder = join(output_folder, 'klusta')

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    # save prb file:
    if probe_file is None:
        si.saveProbeFile(recording, join(output_folder, 'probe.prb'), format='klusta')
        probe_file = join(output_folder, 'probe.prb')
    # save binary file
    if file_name is None:
        file_name = 'recording'
    elif file_name.endswith('.dat'):
        file_name = file_name[file_name.find('.dat')]
    si.writeBinaryDatFormat(recording, join(output_folder, file_name))

    # set up klusta config file
    with open(join(source_dir, 'config_default.prm'), 'r') as f:
        klusta_config = f.readlines()

    klusta_config = ''.join(klusta_config).format(
        join(output_folder, file_name), probe_file, float(recording.getSamplingFrequency()), recording.getNumChannels(), "'float32'",
        threshold_strong_std_factor, threshold_weak_std_factor, "'"+detect_spikes+"'", extract_s_before, extract_s_after,
        n_features_per_channel, pca_n_waveforms_max, num_starting_clusters
    )

    with open(join(output_folder, 'config.prm'), 'w') as f:
        f.writelines(klusta_config)

    print('Running Klusta')
    t_start_proc = time.time()
    cmd = 'klusta {} --overwrite'.format(join(output_folder, 'config.prm'))
    print(cmd)
    retcode = run_command_and_print_output(cmd)
    if retcode != 0:
        raise Exception('Klusta returned a non-zero exit code')
    print('Elapsed time: ', time.time() - t_start_proc)

    sorting = si.KlustaSortingExtractor(join(output_folder, file_name +'.kwik'))

    return sorting



from kbucket import client as kb
import spikeinterface as si
import spikeforest as sf
import mlprocessors as mlpr
import os
import shutil
import random
import string

class MountainSort4(mlpr.Processor):
    NAME='MountainSort4'
    VERSION='4.0.1'
    
    recording_dir=mlpr.Input('Directory of recording',directory=True)
    firings_out=mlpr.Output('Output firings file')
    
    detect_sign=mlpr.IntegerParameter('Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius=mlpr.FloatParameter('Use -1 to include all channels in every neighborhood')
    freq_min=mlpr.FloatParameter(optional=True,default=300,description='Use 0 for no bandpass filtering')
    freq_max=mlpr.FloatParameter(optional=True,default=6000,description='Use 0 for no bandpass filtering')
    whiten=mlpr.BoolParameter(optional=True,default=True,description='Whether to do channel whitening as part of preprocessing')
    clip_size=mlpr.IntegerParameter(optional=True,default=50,description='')
    detect_threshold=mlpr.FloatParameter(optional=True,default=3,description='')
    detect_interval=mlpr.IntegerParameter(optional=True,default=10,description='Minimum number of timepoints between events detected on the same channel')
    noise_overlap_threshold=mlpr.FloatParameter(optional=True,default=0.15,description='Use None for no automated curation')
    
    def run(self):
        recording=si.MdaRecordingExtractor(self.recording_dir)
        num_workers=os.environ.get('NUM_WORKERS',None)
        if num_workers:
            num_workers=int(num_workers)
        sorting=sf.sorters.mountainsort4(
            recording=recording,
            detect_sign=self.detect_sign,
            adjacency_radius=self.adjacency_radius,
            freq_min=self.freq_min,
            freq_max=self.freq_max,
            whiten=self.whiten,
            clip_size=self.clip_size,
            detect_threshold=self.detect_threshold,
            detect_interval=self.detect_interval,
            noise_overlap_threshold=self.noise_overlap_threshold,
            num_workers=num_workers
        )
        si.MdaSortingExtractor.writeSorting(sorting=sorting,save_path=self.firings_out)
    

class IronClust(mlpr.Processor):
    NAME='IronClust'
    VERSION='4.2.6'
    
    recording_dir=mlpr.Input('Directory of recording',directory=True)
    firings_out=mlpr.Output('Output firings file')
    
    detect_sign=mlpr.IntegerParameter('Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius=mlpr.FloatParameter('Use -1 to include all channels in every neighborhood')
    detect_threshold=mlpr.FloatParameter(optional=True,default=3,description='')
    prm_template_name=mlpr.StringParameter(optional=False,description='TODO')
    freq_min=mlpr.FloatParameter(optional=True,default=300,description='Use 0 for no bandpass filtering')
    freq_max=mlpr.FloatParameter(optional=True,default=6000,description='Use 0 for no bandpass filtering')
    merge_thresh=mlpr.FloatParameter(optional=True,default=0.98,description='TODO')
    pc_per_chan=mlpr.IntegerParameter(optional=True,default=3,description='TODO')
    
    def run(self):
        ironclust_src=os.environ.get('IRONCLUST_SRC',None)
        if not ironclust_src:
            raise Exception('Environment variable not set: IRONCLUST_SRC')
        code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
        tmpdir=os.environ.get('TEMPDIR','/tmp')+'/ironclust-tmp-'+code
            
        try:
            recording=si.MdaRecordingExtractor(self.recording_dir)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting=sf.sorters.ironclust(
                recording=recording,
                tmpdir=tmpdir, ## TODO
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                detect_threshold=self.detect_threshold,
                merge_thresh=self.merge_thresh,
                freq_min=self.freq_min,
                freq_max=self.freq_max,
                pc_per_chan=self.pc_per_chan,
                prm_template_name=self.prm_template_name,
                ironclust_src=ironclust_src
            )
            si.MdaSortingExtractor.writeSorting(sorting=sorting,save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            raise
        shutil.rmtree(tmpdir)

#sf.sorters.ironclust(*, recording, tmpdir, detect_sign=-1, adjacency_radius=-1, detect_threshold=5, merge_thresh=0.98, freq_min=300, freq_max=6000, pc_per_chan=3, prm_template_name, ironclust_src=None)
        
Processors=dict(
    MountainSort4=MountainSort4,
    IronClust=IronClust
)
        
def sf_sort_recording(sorter,recording):
    dsdir=recording['directory']
    sorting_params=sorter['params']
    processor_name=sorter['processor_name']
    if processor_name in Processors:
        SS=Processors[processor_name]
    else:
        raise Exception('No such sorter: '+processor_name)
        
    outputs=SS.execute(
        recording_dir=dsdir,
        firings_out=dict(ext='.mda'),
        **sorting_params
    ).outputs
    firings_out=kb.saveFile(outputs['firings_out'])
    result=dict(
        recording_name=recording['name'],
        study_name=recording['study'],
        sorter_name=sorter['name'],
        recording_dir=dsdir,
        firings_true=dsdir+'/firings_true.mda',
        sorting_params=sorting_params,
        sorting_processor_name=SS.NAME,
        sorting_processor_version=SS.VERSION,
        firings=firings_out
    )
    result['summary']=sf.summarizeSorting(result)
    result['comparison_with_truth']=sf.compareWithTruth(result)
    
    return result



import spikeinterface as si
import spiketoolkit as st
import spikewidgets as sw
import json
from PIL import Image
import os
from copy import deepcopy
from kbucket import client as kb
from pairio import client as pa
import mlprocessors as mlpr
from matplotlib import pyplot as plt

def sf_summarize_recording(recording):
  ret=deepcopy(recording)
  ret['computed_info']=compute_recording_info(recording)  
  firings_true_path=recording['directory']+'/firings_true.mda'
  if kb.findFile(firings_true_path):
    ret['ground_truth']=firings_true_path
  ret['plots']=dict(
    timeseries=create_timeseries_plot(recording)
  )
  if ret['ground_truth']:
    ret['plots']['waveforms_true']=create_waveforms_plot(recording,ret['ground_truth'])
  return ret

def read_json_file(fname):
  with open(fname) as f:
    return json.load(f)
  
def write_json_file(fname,obj):
  with open(fname, 'w') as f:
    json.dump(obj, f)
    
def save_plot(fname,quality=40):
    plt.savefig(fname+'.png')
    plt.close()
    im=Image.open(fname+'.png').convert('RGB')
    os.remove(fname+'.png')
    im.save(fname,quality=quality)

# A MountainLab processor for generating the summary info for a recording
class ComputeRecordingInfo(mlpr.Processor):
  NAME='ComputeRecordingInfo'
  VERSION='0.1.0'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  json_out=mlpr.Output('Info in .json file')
    
  def run(self):
    ret={}
    recording=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=False)
    ret['samplerate']=recording.getSamplingFrequency()
    ret['num_channels']=len(recording.getChannelIds())
    ret['duration_sec']=recording.getNumFrames()/ret['samplerate']
    write_json_file(self.json_out,ret)
  
def compute_recording_info(recording):
  out=ComputeRecordingInfo.execute(recording_dir=recording['directory'],json_out={'ext':'.json'}).outputs['json_out']
  kb.saveFile(out)
  return read_json_file(kb.realizeFile(out))

# A MountainLab processor for generating a plot of a portion of the timeseries
class CreateTimeseriesPlot(mlpr.Processor):
  NAME='CreateTimeseriesPlot'
  VERSION='0.1.6'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  jpg_out=mlpr.Output('The plot as a .jpg file')
  
  def run(self):
    R0=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=False)
    R=st.filters.bandpass_filter(recording=R0,freq_min=300,freq_max=6000)
    N=R.getNumFrames()
    N2=int(N/2)
    channels=R.getChannelIds()
    if len(channels)>20: channels=channels[0:20]
    sw.TimeseriesWidget(recording=R,trange=[N2-4000,N2+0],channels=channels,width=12,height=5).plot()
    save_plot(self.jpg_out)
    
def create_timeseries_plot(recording):
  out=CreateTimeseriesPlot.execute(recording_dir=recording['directory'],jpg_out={'ext':'.jpg'}).outputs['jpg_out']
  kb.saveFile(out)
  return 'sha1://'+kb.computeFileSha1(out)+'/timeseries.jpg'

# A MountainLab processor for generating a plot of a portion of the timeseries
class CreateWaveformsPlot(mlpr.Processor):
  NAME='CreateWaveformsPlot'
  VERSION='0.1.0'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  firings=mlpr.Input(description='Firings file')
  jpg_out=mlpr.Output('The plot as a .jpg file')
  
  def run(self):
    R0=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=True)
    R=st.filters.bandpass_filter(recording=R0,freq_min=300,freq_max=6000)
    S=si.MdaSortingExtractor(firings_file=self.firings)
    channels=R.getChannelIds()
    if len(channels)>20:
      channels=channels[0:20]
    units=S.getUnitIds()
    if len(units)>20:
      units=units[::int(len(units)/20)]
    sw.UnitWaveformsWidget(recording=R,sorting=S,channels=channels,unit_ids=units).plot()
    save_plot(self.jpg_out)
    
def create_waveforms_plot(recording,firings):
  out=CreateWaveformsPlot.execute(recording_dir=recording['directory'],firings=firings,jpg_out={'ext':'.jpg'}).outputs['jpg_out']
  kb.saveFile(out)
  return 'sha1://'+kb.computeFileSha1(out)+'/waveforms.jpg'




from .process_datasets import download_datasets, process_datasets, clear_dataset_results, assemble_dataset_results, try_process_dataset
from .sort_datasets import sort_datasets, clear_sorting_results, assemble_sorting_results, try_sort_dataset


#from .sfstudy import SFStudy
#from . import studies
from .sortdataset import sortDataset, MountainSort4
from .summarizesorting import summarizeSorting
from .comparewithtruth import compareWithTruth
from .kbucketconfig import kbucketConfigLocal, kbucketConfigRemote
from .sfdata import SFData
from . import batch
from . import sorters
from . import sf_batch


#!/usr/bin/env python

from kbucket import client as kb
from pairio import client as pa
import spikeforest as sf
import spikeinterface as si
import spikewidgets as sw
import spiketoolkit as st
from copy import deepcopy
import json
import mlprocessors as mlpr
from matplotlib import pyplot as plt
from PIL import Image
import os
import random
import string

default_run_code='0'

def clear_dataset_results(in_process_only,run_code=default_run_code):
  print('Loading tasks...')
  tasks=load_tasks(run_code=run_code)

  print('Checking tasks...')
  for i,task in enumerate(tasks):
    task.clearResults(in_process_only=in_process_only)
  
  print('Done.')

def download_datasets(run_code=default_run_code):
  tasks=load_tasks(run_code=run_code)

  for i,task in enumerate(tasks):
    ds=task.dataset()
    print('Download task {} of {}: {}'.format(i+1,len(tasks),ds['name']))
    dsdir=ds['directory']
    kb.realizeFile(dsdir+'/raw.mda')

def process_datasets(run_code=default_run_code):
  tasks=load_tasks(run_code=run_code)

  for i,task in enumerate(tasks):
    ds=task.dataset()
    print('Processing task {} of {}: {}'.format(i+1,len(tasks),ds['name']))
    task.execute()
    
def try_process_dataset(study_name,dataset_name,run_code=default_run_code):
  tasks=load_tasks(run_code=run_code)
  for task in tasks:
    ds=task.dataset()
    if ds['study']==study_name:
      if ds['name']==dataset_name:
        result=task.run()
        return result
  raise Exception('Not found')

def assemble_dataset_results(run_code=default_run_code):
  tasks=load_tasks(run_code=run_code)

  results=[]
  for i,task in enumerate(tasks):
    print('Loading result for task {} of {}: {}/{}'.format(i+1,len(tasks),task.dataset()['study'],task.dataset()['name']))
    result=task.loadResult()
    if not result:
      raise Exception('Unable to load result for task.')
    results.append(result)

  key1=dict(name='spikeforest_studies')
  key2=dict(name='spikeforest_studies_processed')
  print('Saving results to... key={}'.format(json.dumps(key2)))
  obj=kb.loadObject(
    key=key1,
    share_ids=['spikeforest.spikeforest1']
  )
  obj['datasets']=results;
  datasets=obj['datasets']
  for ds in datasets:
      print(ds['study'],ds['name'])
  kb.saveObject(obj,key=key2)


def load_tasks(run_code):
  obj=kb.loadObject(
    key=dict(name='spikeforest_studies'),
    share_ids=['spikeforest.spikeforest1']
  )
  
  datasets=obj['datasets']
  tasks=[]
  for i,ds in enumerate(datasets):
    key=dict(
            script='process_datasets',
            study_name=ds['study'],
            dataset_name=ds['name'],
            run_code=run_code
        )
    tasks.append(
        ProcessDatasetTask(key,ds)
    )
  return tasks

class ProcessDatasetTask():
  def __init__(self,key,dataset):
    self._key=key
    self._dataset=dataset
    self._code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
    
  def dataset(self):
    return self._dataset
    
  def clearResults(self,*,in_process_only):
    val=pa.get(self._key)
    if val:
      if (not in_process_only) or (val.startswith('in-process')) or (val.startswith('error')):
        print('Clearing results for: '+self._key['dataset_name'])
        pa.set(key=self._key,value=None)

  def execute(self):
    print('Dataset: '+self._dataset['name'])
    val=pa.get(self._key)
    if val:
      if val.startswith('in-process'):
        print('In progress, skipping...')
      else:
        print('Completed, skipping...')
      return
    if not pa.set(self._key,'in-process-'+self._code,overwrite=False):
      print('Problem setting in-process value skipping...')
      return
    try:
      result=self.run()
    except:
      if pa.get(self._key)=='in-process-'+self._code:
        pa.set(self._key,'error')
      else:
        print('Unexpected: not setting error value because existing value does not match')
      raise
    if pa.get(self._key)=='in-process-'+self._code:
      print('Saving result object.')
      kb.saveObject(key=self._key,object=result)
    else:
      print('Unexpected: not setting result because existing value does not match.')

  def loadResult(self):
    val=pa.get(self._key)
    if val:
      if val.startswith('in-process'):
        print('Inable to load result... it is in process.')
        return None
      else:
        return kb.loadObject(key=self._key)
    else:
      return None
  
  def run(self):
    return process_dataset(self._dataset)

def read_json_file(fname):
  with open(fname) as f:
    return json.load(f)
  
def write_json_file(fname,obj):
  with open(fname, 'w') as f:
    json.dump(obj, f)
    
def save_plot(fname,quality=40):
    plt.savefig(fname+'.png')
    plt.close()
    im=Image.open(fname+'.png').convert('RGB')
    os.remove(fname+'.png')
    im.save(fname,quality=quality)

# A MountainLab processor for generating the summary info for a dataset
class ComputeDatasetInfo(mlpr.Processor):
  NAME='ComputeDatasetInfo'
  VERSION='0.1.0'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  json_out=mlpr.Output('Info in .json file')
    
  def run(self):
    ret={}
    recording=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=False)
    ret['samplerate']=recording.getSamplingFrequency()
    ret['num_channels']=len(recording.getChannelIds())
    ret['duration_sec']=recording.getNumFrames()/ret['samplerate']
    write_json_file(self.json_out,ret)
  
def compute_dataset_info(dataset):
  out=ComputeDatasetInfo.execute(recording_dir=dataset['directory'],json_out={'ext':'.json'}).outputs['json_out']
  kb.saveFile(out)
  return read_json_file(kb.realizeFile(out))

# A MountainLab processor for generating a plot of a portion of the timeseries
class CreateTimeseriesPlot(mlpr.Processor):
  NAME='CreateTimeseriesPlot'
  VERSION='0.1.6'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  jpg_out=mlpr.Output('The plot as a .jpg file')
  
  def run(self):
    R0=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=False)
    R=st.filters.bandpass_filter(recording=R0,freq_min=300,freq_max=6000)
    N=R.getNumFrames()
    N2=int(N/2)
    channels=R.getChannelIds()
    if len(channels)>20: channels=channels[0:20]
    sw.TimeseriesWidget(recording=R,trange=[N2-4000,N2+0],channels=channels,width=12,height=5).plot()
    save_plot(self.jpg_out)
    
def create_timeseries_plot(dataset):
  out=CreateTimeseriesPlot.execute(recording_dir=dataset['directory'],jpg_out={'ext':'.jpg'}).outputs['jpg_out']
  kb.saveFile(out)
  return 'sha1://'+kb.computeFileSha1(out)+'/timeseries.jpg'

# A MountainLab processor for generating a plot of a portion of the timeseries
class CreateWaveformsPlot(mlpr.Processor):
  NAME='CreateWaveformsPlot'
  VERSION='0.1.0'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  firings=mlpr.Input(description='Firings file')
  jpg_out=mlpr.Output('The plot as a .jpg file')
  
  def run(self):
    R0=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=True)
    R=st.filters.bandpass_filter(recording=R0,freq_min=300,freq_max=6000)
    S=si.MdaSortingExtractor(firings_file=self.firings)
    channels=R.getChannelIds()
    if len(channels)>20:
      channels=channels[0:20]
    units=S.getUnitIds()
    if len(units)>20:
      units=units[::int(len(units)/20)]
    sw.UnitWaveformsWidget(recording=R,sorting=S,channels=channels,unit_ids=units).plot()
    save_plot(self.jpg_out)
    
def create_waveforms_plot(dataset,firings):
  out=CreateWaveformsPlot.execute(recording_dir=dataset['directory'],firings=firings,jpg_out={'ext':'.jpg'}).outputs['jpg_out']
  kb.saveFile(out)
  return 'sha1://'+kb.computeFileSha1(out)+'/waveforms.jpg'

def process_dataset(dataset):
  ret=deepcopy(dataset)
  ret['computed_info']=compute_dataset_info(dataset)  
  firings_true_path=dataset['directory']+'/firings_true.mda'
  if kb.findFile(firings_true_path):
    ret['ground_truth']=firings_true_path
  ret['plots']=dict(
    timeseries=create_timeseries_plot(dataset)
  )
  if ret['ground_truth']:
    ret['plots']['waveforms_true']=create_waveforms_plot(dataset,ret['ground_truth'])
  return ret

    


import spikeforest as sf
from kbucket import client as kb
from pairio import client as pa
import json
import os
import random
import string
from .sf_summarize_recording import sf_summarize_recording
from .sf_sort_recording import sf_sort_recording

def login(config):
    password=os.environ.get('SPIKEFOREST_PASSWORD',None)
    if not password:
        raise Exception('Environment variable not set: SPIKEFOREST_PASSWORD')
    sf.kbucketConfigRemote(share_id=config['share_id'],write=True,password=os.environ['SPIKEFOREST_PASSWORD'])

def sf_batch_prepare(config,*,clear_all=False):
    login(config)
    study_obj=kb.loadObject(key=dict(name='spikeforest_recordings'))
    recordings=select_recordings(study_obj,config)
    sorters=config['sorters']
    
    clear_in_process_only=(not clear_all)
    for ds in recordings:
        print('PREPARE: {}/{}'.format(ds['study'],ds['name']))
        print('Downloading raw.mda')
        dsdir=ds['directory']
        kb.realizeFile(dsdir+'/raw.mda')
        
        if config.get('summarize_recordings',None):
            key=dict(
                name='summarize_recording',
                batch_name=config['batch_name'],
                study_name=ds['study'],
                recording_name=ds['name']
            )
            clear_result_for_key(key=key,in_process_only=clear_in_process_only)
        
        for sorter in sorters:
            key=dict(
                name='sort_recording',
                batch_name=config['batch_name'],
                study_name=ds['study'],
                recording_name=ds['name'],
                sorter_name=sorter['name'],
                sorter_params=sorter['params']
            )
            clear_result_for_key(key=key,in_process_only=clear_in_process_only)

def sf_batch_run(config):
    login(config)
    study_obj=kb.loadObject(key=dict(name='spikeforest_recordings'))
    recordings=select_recordings(study_obj,config)
    sorters=config['sorters']
    
    code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
    for i,ds in enumerate(recordings):
        if config.get('summarize_recordings',None):
            key=dict(
                name='summarize_recording',
                batch_name=config['batch_name'],
                study_name=ds['study'],
                recording_name=ds['name']
            )
            if acquire_lock_for_key(key=key,code=code):
                try:
                    print('========= Summarizing recording {}/{}: {}/{}'.format(i,len(recordings),ds['study'],ds['name']))
                    result0=sf_summarize_recording(ds)
                except:
                    if check_consistent_code(key=key,code=code):
                        pa.set(key=key,value='error-'+code)
                    raise
                if check_consistent_code(key=key,code=code):
                    kb.saveObject(key=key,object=result0)
                else:
                    print('Warning: inconsistent code for {}'.format(json.dumps(key)))

        for sorter in sorters:
            key=dict(
                name='sort_recording',
                batch_name=config['batch_name'],
                study_name=ds['study'],
                recording_name=ds['name'],
                sorter_name=sorter['name'],
                sorter_params=sorter['params']
            )
            if acquire_lock_for_key(key=key,code=code):
                try:
                    print('========= Sorting recording {}/{}: {} - {}/{}'.format(i,len(recordings),sorter['name'],ds['study'],ds['name']))
                    result0=sf_sort_recording(sorter,ds)
                except:
                    if check_consistent_code(key=key,code=code):
                        pa.set(key=key,value='error-'+code)
                    raise
                if check_consistent_code(key=key,code=code):
                    kb.saveObject(key=key,object=result0)
                else:
                    print('Warning: inconsistent code for {}'.format(json.dumps(key)))
                
def sf_batch_assemble(config):
    login(config)
    study_obj=kb.loadObject(key=dict(name='spikeforest_recordings'))
    recordings=select_recordings(study_obj,config)
    sorters=config['sorters']
    
    batch_output=dict(
        recordings=recordings,
        sorters=sorters,
        summarize_recording_results=[],
        sorting_results=[]
    )
    for ds in recordings:
        if config.get('summarize_recordings',None):
            key=dict(
                name='summarize_recording',
                batch_name=config['batch_name'],
                study_name=ds['study'],
                recording_name=ds['name']
            )
            result0=kb.loadObject(key=key)
            if not result0:
                raise Exception('Problem loading summarize_recording result {}'.format(json.dumps(key)))
            batch_output['summarize_recording_results'].append(result0)

        for sorter in sorters:
            print('ASSEMBLE: {} {}/{}'.format(sorter['name'],ds['study'],ds['name']))
            key=dict(
                name='sort_recording',
                batch_name=config['batch_name'],
                study_name=ds['study'],
                recording_name=ds['name'],
                sorter_name=sorter['name'],
                sorter_params=sorter['params']
            )
            result0=kb.loadObject(key=key)
            if not result0:
                raise Exception('Problem loading sort_recording result {}'.format(json.dumps(key)))
            batch_output['sorting_results'].append(result0)
            
    kb.saveObject(
        key=dict(
            batch_name=config['batch_name']
        ),
        object=batch_output
    )


def select_recordings(study_obj,config):
    recordings=[]
    for ds in study_obj['recordings']:
        if (not config['studies']) or (ds['study'] in config['studies']):
            if (not config['recordings']) or (ds['name'] in config['recordings']):
                recordings.append(ds)
    return recordings

def clear_result_for_key(*,key,in_process_only=False):
    val=pa.get(key=key)
    if val:
        if in_process_only:
            do_clear=((val.startswith('in-process')) or (val.startswith('error')))
        else:
            do_clear=True
        if do_clear:
            print('Clearing results for: {}'+json.dumps(key))
            pa.set(key=key,value=None)

def acquire_lock_for_key(*,key,code):
    val=pa.get(key=key)
    if val:
        if val.startswith('in-process'):
            return False
        if val.startswith('error'):
            return False
        return False
    if not pa.set(key,'in-process-'+code,overwrite=False):
        return False
    return True

def check_consistent_code(*,key,code):
    val=pa.get(key=key)
    if not val:
        return False
    return (val=='in-process-'+code)


#!/usr/bin/env python

import argparse
import spikeforest as sf
import os
import json

def read_json_file(fname):
    with open(fname) as f:
        return json.load(f)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description = 'Sort the SpikeForest recordings')
    parser.add_argument('command',help='clear, prepare, run, assemble')
    parser.add_argument('config_file',help='Name of config .json file')
    args = parser.parse_args()

    spikeforest_password=os.environ.get('SPIKEFOREST_PASSWORD','')
    if not spikeforest_password:
      raise Exception('Environment variable not set: SPIKEFOREST_PASSWORD')

    command=args.command
    config_fname=args.config_file
    
    config=read_json_file(config_fname)
    if command=='clear':
      sf.sf_batch.sf_batch_prepare(config,clear_all=True)
    elif command=='prepare':
      sf.sf_batch.sf_batch_prepare(config)
    elif command=='run':
      sf.sf_batch.sf_batch_run(config)
    elif command=='assemble':
      sf.sf_batch.sf_batch_assemble(config)
    else:
      raise Exception('Unrecognized command: '+command)


from kbucket import client as kb
import spikeinterface as si
import spikeforest as sf
import tempfile
import random
import mlprocessors as mlpr
import os

class MountainSort4(mlpr.Processor):
    NAME='MountainSort4'
    VERSION='4.0.1'
    
    dataset_dir=mlpr.Input('Directory of dataset',directory=True)
    firings_out=mlpr.Output('Output firings file')
    
    detect_sign=mlpr.IntegerParameter('Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius=mlpr.FloatParameter('Use -1 to include all channels in every neighborhood')
    freq_min=mlpr.FloatParameter(optional=True,default=300,description='Use 0 for no bandpass filtering')
    freq_max=mlpr.FloatParameter(optional=True,default=6000,description='Use 0 for no bandpass filtering')
    whiten=mlpr.BoolParameter(optional=True,default=True,description='Whether to do channel whitening as part of preprocessing')
    clip_size=mlpr.IntegerParameter(optional=True,default=50,description='')
    detect_threshold=mlpr.FloatParameter(optional=True,default=3,description='')
    detect_interval=mlpr.IntegerParameter(optional=True,default=10,description='Minimum number of timepoints between events detected on the same channel')
    noise_overlap_threshold=mlpr.FloatParameter(optional=True,default=0.15,description='Use None for no automated curation')
    
    def run(self):
        recording=si.MdaRecordingExtractor(self.dataset_dir)
        num_workers=int(os.environ.get('NUM_WORKERS',-1))
        if num_workers<=0: num_workers=None
        sorting=sf.sorters.mountainsort4(
            recording=recording,
            detect_sign=self.detect_sign,
            adjacency_radius=self.adjacency_radius,
            freq_min=self.freq_min,
            freq_max=self.freq_max,
            whiten=self.whiten,
            clip_size=self.clip_size,
            detect_threshold=self.detect_threshold,
            detect_interval=self.detect_interval,
            noise_overlap_threshold=self.noise_overlap_threshold,
            num_workers=num_workers
        )
        si.MdaSortingExtractor.writeSorting(sorting=sorting,save_path=self.firings_out)

def sortDataset(
    *,
    sorter,
    dataset,
    _force_run=False
):
    dsdir=dataset['directory']
    sorting_params=sorter['params']
    sorting_processor=sorter['processor']
    outputs=sorting_processor.execute(dataset_dir=dsdir,firings_out=dict(ext='.mda'),**sorting_params,_force_run=_force_run).outputs
    result=dict(
        dataset_name=dataset['name'],
        dataset_dir=dsdir,
        firings_true=dsdir+'/firings_true.mda',
        sorting_params=sorting_params,
        sorting_processor_name=sorting_processor.NAME,
        sorting_processor_version=sorting_processor.VERSION,
        firings=outputs['firings_out']
    )
    kb.saveFile(outputs['firings_out'])
    return result

def _create_temporary_fname(ext):
    return tempfile.gettempdir()+'/tmp_'+''.join(random.choices('abcdefghijklmnopqrstuvwxyz', k=10))+ext



import spikeinterface as si

import os
import time
import numpy as np
from os.path import join
from spiketoolkit.sorters.tools import run_command_and_print_output

def kilosort(
        recording,
        kilosort_path=None,
        npy_matlab_path=None,
        output_folder=None,
        useGPU=False,
        probe_file=None,
        file_name=None,
        spike_thresh=5,
        electrode_dimensions=None
    ):
    if kilosort_path is None:
        kilosort_path = os.getenv('KILOSORT_PATH', None)
    if npy_matlab_path is None:
        npy_matlab_path = os.getenv('NPY_MATLAB_PATH', None)
    if not os.path.isfile(join(kilosort_path, 'preprocessData.m')) \
            or not os.path.isfile(join(npy_matlab_path, 'readNPY.m')):
        raise ModuleNotFoundError("\nTo use KiloSort, install KiloSort and npy-matlab from sources: \n\n"
                                  "\ngit clone https://github.com/cortex-lab/KiloSort\n"
                                  "\ngit clone https://github.com/kwikteam/npy-matlab\n"
                                  "and provide the installation path with the 'kilosort_path' and "
                                  "'npy_matlab_path' arguments or by setting the KILOSORT_PATH and NPY_MATLAB_PATH"
                                  "environment variables.\n+n"
                                  "\nMore information on KiloSort at: "
                                  "\nhttps://github.com/cortex-lab/KiloSort")

    source_dir = os.path.dirname(os.path.realpath(__file__))

    if output_folder is None:
        output_folder = os.path.abspath('kilosort')
    else:
        output_folder = os.path.abspath(join(output_folder, 'kilosort'))

    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)

    kilosort_path = os.path.abspath(kilosort_path)
    npy_matlab_path = os.path.abspath(npy_matlab_path)

    if probe_file is not None:
        si.loadProbeFile(recording, probe_file)

    # save binary file
    if file_name is None:
        file_name = 'recording'
    elif file_name.endswith('.dat'):
        file_name = file_name[file_name.find('.dat')]
    si.writeBinaryDatFormat(recording, join(output_folder, file_name))

    # set up kilosort config files and run kilosort on data
    with open(join(source_dir, 'kilosort_master.txt'), 'r') as f:
        kilosort_master = f.readlines()
    with open(join(source_dir, 'kilosort_config.txt'), 'r') as f:
        kilosort_config = f.readlines()
    with open(join(source_dir, 'kilosort_channelmap.txt'), 'r') as f:
        kilosort_channelmap = f.readlines()

    nchan = recording.getNumChannels()
    dat_file = file_name +'.dat'
    kilo_thresh = spike_thresh
    Nfilt = (nchan // 32) * 32 * 4
    if Nfilt == 0:
        Nfilt = 64
    nsamples = 128 * 1024 + 32

    if useGPU:
        ug = 1
    else:
        ug = 0

    kilosort_master = ''.join(kilosort_master).format(
        ug, kilosort_path, npy_matlab_path, output_folder, join(output_folder, 'results')
    )
    kilosort_config = ''.join(kilosort_config).format(
        nchan, nchan, recording.getSamplingFrequency(), dat_file , Nfilt, nsamples, kilo_thresh
    )
    if 'location' in recording.getChannelPropertyNames():
        positions = np.array([recording.getChannelProperty(chan, 'location') for chan in range(nchan)])
        if electrode_dimensions is None:
            kilosort_channelmap = ''.join(kilosort_channelmap
                                          ).format(nchan,
                                                   list(positions[:, 0]),
                                                   list(positions[:, 1]),
                                                   'ones(1, Nchannels)',
                                                   recording.getSamplingFrequency())
        elif len(electrode_dimensions)==2:
            kilosort_channelmap = ''.join(kilosort_channelmap
                                          ).format(nchan,
                                                   list(positions[:, electrode_dimensions[0]]),
                                                   list(positions[:, electrode_dimensions[1]]),
                                                   'ones(1, Nchannels)',
                                                   recording.getSamplingFrequency())
        else:
            raise Exception("Electrode dimension should bi a list of len 2")


    else:
        raise Exception("'location' information is needed. Provide a probe information with a 'probe_file'")

    for fname, value in zip(['kilosort_master.m', 'kilosort_config.m',
                             'kilosort_channelmap.m'],
                            [kilosort_master, kilosort_config,
                             kilosort_channelmap]):
        with open(join(output_folder, fname), 'w') as f:
            f.writelines(value)

    # start sorting with kilosort
    print('Running KiloSort')
    cwd = os.getcwd()
    t_start_proc = time.time()
    os.chdir(output_folder)
    print(os.getcwd())
    cmd = 'matlab -nosplash -nodisplay -r "run kilosort_master.m; quit;"'
    print(cmd)
    retcode = run_command_and_print_output(cmd)

    if retcode != 0:
        raise Exception('KiloSort returned a non-zero exit code')
    print('Elapsed time: ', time.time() - t_start_proc)

    sorting = si.KiloSortSortingExtractor(join(output_folder, 'results'))
    os.chdir(cwd)
    return sorting



from mountainlab_pytools import mlproc as mlp
import spikeinterface as si

class SFStudy():
  def __init__(self):
    self._datasets=dict()

  def getDatasetIDs(self):
    return sorted(self._datasets.keys())

  def addDataset(self,*,dataset_id,dataset):
    self._datasets[dataset_id]=dataset

  def getDataset(self,dataset_id):
    return self._datasets[dataset_id]

  def loadDatasetsFromKBucket(self,*,kbucket_path,dataset_id_prefix=''):
    D=mlp.readDir(kbucket_path)
    for name in D['dirs']:
      DS=SFDataset(kbucket_path=kbucket_path+'/'+name)
      self.addDataset(dataset_id=dataset_id_prefix+name,dataset=DS)

class SFDataset():
  def __init__(self,kbucket_path):
    self._kbucket_path=kbucket_path
  
  def getRecording(self,download=True):
    ret=si.MdaRecordingExtractor(dataset_directory=self._kbucket_path,download=download)
    return ret

  def getSortingTrue(self):
    D2=mlp.readDir(self._kbucket_path)
    if 'firings_true.mda' in D2['files']:
      ret=si.MdaSortingExtractor(firings_file=self._kbucket_path+'/firings_true.mda')
    else:
      ret=None
    return ret



from matplotlib import pyplot as plt
from PIL import Image
import mlprocessors as mlpr
from kbucket import client as kb
import spikeinterface as si
import spikewidgets as sw
import os

def summarizeSorting(result):
    ret={'plots':{}}
    unit_waveforms=PlotUnitWaveforms.execute(recording_dir=result['recording_dir'],firings=result['firings'],plot_out={'ext':'.jpg'}).outputs['plot_out']
    unit_waveforms=kb.saveFile(unit_waveforms,basename='unit_waveforms.jpg')
    ret['plots']['unit_waveforms']=unit_waveforms

    autocorrelograms=PlotAutoCorrelograms.execute(recording_dir=result['recording_dir'],firings=result['firings'],plot_out={'ext':'.jpg'}).outputs['plot_out']
    autocorrelograms=kb.saveFile(autocorrelograms,basename='autocorrelograms.jpg')
    ret['plots']['autocorrelograms']=autocorrelograms

    return ret

class PlotUnitWaveforms(mlpr.Processor):
    recording_dir=mlpr.Input(directory=True,description='Recording directory')
    firings=mlpr.Input('Firings file (sorting)')
    plot_out=mlpr.Output('Plot as .jpg image file')
    
    def run(self):
        recording=si.MdaRecordingExtractor(dataset_directory=self.recording_dir)
        sorting=si.MdaSortingExtractor(firings_file=self.firings)
        sw.UnitWaveformsWidget(recording=recording,sorting=sorting).plot()
        fname=save_plot(self.plot_out)

class PlotAutoCorrelograms(mlpr.Processor):
    NAME='spikeforest.PlotAutoCorrelograms'
    VERSION='0.1.0'
    recording_dir=mlpr.Input(directory=True,description='Recording directory')
    firings=mlpr.Input('Firings file (sorting)')
    plot_out=mlpr.Output('Plot as .jpg image file')
    
    def run(self):
        recording=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=False)
        sorting=si.MdaSortingExtractor(firings_file=self.firings)
        sw.CrossCorrelogramsWidget(samplerate=recording.getSamplingFrequency(),sorting=sorting).plot()
        fname=save_plot(self.plot_out)

def save_plot(fname,quality=40):
    plt.savefig(fname+'.png')
    plt.close()
    im=Image.open(fname+'.png').convert('RGB')
    os.remove(fname+'.png')
    im.save(fname,quality=quality)


import spikeforest as sf
from kbucket import client as kb
from pairio import client as pa
import random
import string
import json
from copy import deepcopy

default_run_code='0'

def sort_datasets(*,study_name,run_code=default_run_code):
    tasks=load_tasks(study_name=study_name)
    for task in tasks:
        task.execute()
        
def clear_sorting_results(*,study_name,in_process_only):
    tasks=load_tasks(study_name=study_name)
    for task in tasks:
        task.clearResults(in_process_only=in_process_only)
        
def assemble_sorting_results(study_name,run_code=default_run_code):
    tasks = load_tasks(study_name=study_name,run_code=run_code)

    results = []
    for (i, task) in enumerate(tasks):
        print('Loading result for task {} of {}: {} {}/{}'.format(
            i + 1, len(tasks),
            task.sorter()['name'],
            task.dataset()['study'], 
            task.dataset()['name']
        ))
        result = task.loadResult()
        if not result:
            raise Exception('Unable to load result for task.')
        results.append(result)

    key = dict(name='spikeforest_sorting_results',study=study_name)
    print('Saving results to... key={}'.format(json.dumps(key)))
    obj = dict()
    obj['sorting_results'] = results
    kb.saveObject(obj, key=key)
    
def try_sort_dataset(*,study_name,dataset_name,sorter_name,run_code=default_run_code):
    tasks=load_tasks(study_name=study_name,run_code=run_code)
    for task in tasks:
        ds=task.dataset()
        so=task.sorter()
        if ds['study']==study_name:
            if ds['name']==dataset_name:
                if so['name']==sorter_name:
                    result=task.run()
                    return result
    raise Exception('Not found')

def load_sorters():
    sorters=[]

    ms4_p1=dict(
        detect_sign=-1,
        adjacency_radius=-1,
        detect_threshold=3
    )
    ms4_p2=dict(
        detect_sign=-1,
        adjacency_radius=-1,
        detect_threshold=5
    )
    
    sorters.append(dict(
        name='MountainSort4-p1',
        processor=sf.MountainSort4,
        params=ms4_p1
    ))
    sorters.append(dict(
        name='MountainSort4-p2',
        processor=sf.MountainSort4,
        params=ms4_p2
    ))
    return sorters

def sort_dataset(sorter,dataset):
    ret=dict()
    ret['sorter']=dict(
        name=sorter['name'],
        processor_name=sorter['processor'].NAME,
        processor_version=sorter['processor'].VERSION,
        sorting_params=sorter['params']
    )
    ret['dataset']=deepcopy(dataset)
    result=sf.sortDataset(
        sorter = sorter,
        dataset = dataset
    )
    result['comparison_with_truth'] = sf.compareWithTruth(result)
    result['summary'] = sf.summarizeSorting(result)
    ret['result']=result
    return ret

class SortDatasetTask():
    def __init__(self,key,dataset,sorter):
        self._key=key
        self._dataset=dataset
        self._sorter=sorter
        self._code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
    
    def dataset(self):
        return self._dataset

    def sorter(self):
        return self._sorter

    def clearResults(self,*,in_process_only):
        val=pa.get(self._key)
        if val:
            if (not in_process_only) or (val.startswith('in-process')) or (val.startswith('error')):
                print('Clearing results for: '+self._key['dataset_name'])
                pa.set(key=self._key,value=None)

    def execute(self):
        print(self._sorter['name']+' '+self._dataset['name'])
        val=pa.get(self._key)
        if val:
            if val.startswith('in-process'):
                print('In progress, skipping...')
            else:
                print('Completed, skipping...')
                return
        if not pa.set(self._key,'in-process-'+self._code,overwrite=False):
            print('Problem setting in-process value skipping...')
            return
        try:
            result=self.run()
        except:
            if pa.get(self._key)=='in-process-'+self._code:
                pa.set(self._key,'error')
            else:
                print('Unexpected: not setting error value because existing value does not match')
            raise
        if pa.get(self._key)=='in-process-'+self._code:
            print('Saving result object.')
            kb.saveObject(key=self._key,object=result)
        else:
            print('Unexpected: not setting result because existing value does not match.')

    def loadResult(self):
        val=pa.get(self._key)
        if val:
            if val.startswith('in-process'):
                print('Unable to load result... it is in process.')
                return None
            else:
                return kb.loadObject(key=self._key)
        else:
            return None

    def run(self):
        return sort_dataset(self._sorter, self._dataset)
    
def load_tasks(study_name,run_code=default_run_code):
    obj=kb.loadObject(
        key=dict(name='spikeforest_studies_processed'),
        share_ids=['spikeforest.spikeforest1']
    )
    
    datasets=obj['datasets']
    
    tasks=[]
    sorters=load_sorters()
    for ds in datasets:
        if ds['study']==study_name:
            for sorter in sorters:
                key=dict(
                    script='sort_datasets',
                    study_name=ds['study'],
                    dataset_name=ds['name'],
                    sorter_name=sorter['name'],
                    run_code=run_code
                )
                task=SortDatasetTask(dataset=ds,sorter=sorter,key=key)
                tasks.append(task)
    return tasks




import setuptools

pkg_name="spikeforest"

setuptools.setup(
    name=pkg_name,
    version="0.3.0",
    author="Jeremy Magland",
    author_email="jmagland@flatironinstitute.org",
    description="",
    url="https://github.com/magland/spikeforest",
    packages=setuptools.find_packages(),
    install_requires=[
        'pybind11',
        'ml_ms4alg',
        'numpy',
        'matplotlib',
        'spikeinterface',
        'spikewidgets',
        'spiketoolkit',
        'kbucket',
        'pillow',
        'mlprocessors',
        'pandas'
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    )
)



from kbucket import client as kb
from PIL import Image
import json
import pandas as pd
import spikeinterface as si

def kb_read_text_file(fname):
    fname=kb.realizeFile(fname)
    with open(fname,'r') as f:
        return f.read()
    
def kb_read_json_file(fname):
    fname=kb.realizeFile(fname)
    with open(fname,'r') as f:
        return json.load(f)

class SFSortingResult():
    def __init__(self,obj,recording):
        self._obj=obj
        self._recording=recording
    def getObject(self):
        return self._obj
    def recording(self):
        return self._recording
    def sorterName(self):
        return self._obj['sorter_name']
    def plotNames(self):
        plots=self._obj['summary'].get('plots',dict())
        return list(plots.keys())
    def sorting(self):
        return si.MdaSortingExtractor(firings_file=self._obj['firings'])
    def plot(self,name,format='image'):
        plots=self._obj['summary'].get('plots',dict())
        url=plots[name]
        if format=='url':
            return url
        else:
            path=kb.realizeFile(url)
            if format=='image':
                return Image.open(path)
            elif format=='path':
                return path
            else:
                raise Exception('Invalid format: '+format)
    def comparisonWithTruth(self,*,format='dataframe'):
        A=self._obj['comparison_with_truth']
        if format=='html':
            return kb_read_text_file(A['html'])
        else:
            B=kb_read_json_file(A['json'])
            if format=='json':
                return B
            elif format=='dataframe':
                return pd.DataFrame(B).transpose()
            else:
                raise Exception('Invalid format: '+format)

class SFRecording():
    def __init__(self,obj,study):
        self._obj=obj
        self._sorting_result_names=[]
        self._sorting_results_by_name=dict()
        self._study=study
    def getObject(self):
        return self._obj
    def study(self):
        return self._study
    def name(self):
        return self._obj['name']
    def description(self):
        return self._obj['description']
    def directory(self):
        return self._obj['directory']
    def recording(self,download=False):
        return si.MdaRecordingExtractor(dataset_directory=self.directory(),download=download)
    def sortingTrue(self):
        return si.MdaSortingExtractor(firings_file=self.directory()+'/firings_true.mda')
    def plotNames(self):
        plots=self._obj.get('plots',dict())
        return list(plots.keys())
    def plot(self,name,format='image'):
        plots=self._obj.get('plots',dict())
        url=plots[name]
        if format=='url':
            return url
        else:
            path=kb.realizeFile(url)
            if format=='image':
                return Image.open(path)
            elif format=='path':
                return path
            else:
                raise Exception('Invalid format: '+format)
    def addSortingResult(self,obj):
        sorter_name=obj['sorter_name']
        if sorter_name in self._sorting_results_by_name:
            print('Sorting result already in recording: {}'.format(sorter_name))
        else:
            R=SFSortingResult(obj,self)
            self._sorting_result_names.append(sorter_name)
            self._sorting_results_by_name[sorter_name]=R
    def sortingResultNames(self):
        return self._sorting_result_names
    def sortingResult(self,name):
        return self._sorting_results_by_name[name]

class SFStudy():
    def __init__(self,obj):
        self._obj=obj
        self._recordings_by_name=dict()
        self._recording_names=[]
    def getObject(self):
        return self._obj
    def name(self):
        return self._obj['name']
    def description(self):
        return self._obj['description']
    def addRecording(self,obj):
        name=obj['name']
        if name in self._recordings_by_name:
            print('Recording already in study: '+name)
        else:
            self._recording_names.append(name)
            D=SFRecording(obj,self)
            self._recordings_by_name[name]=D
    def recordingNames(self):
        return self._recording_names
    def recording(self,name):
        return self._recordings_by_name[name]
        

class SFData():
    def __init__(self):
        self._studies_by_name=dict()
        self._study_names=[]
    def loadStudies(self,*,key=None):
        if key is None:
            key=dict(name='spikeforest_studies_processed')
        obj=kb.loadObject(key=key)
        studies=obj['studies']
        for study in studies:
            name=study['name']
            if name in self._studies_by_name:
                print('Study already loaded: '+name)
            else:
                self._study_names.append(study['name'])
                S=SFStudy(study)
                self._studies_by_name[name]=S
        recordings=obj['recordings']
        for ds in recordings:
            study=ds['study']
            self._studies_by_name[study].addRecording(ds)
    def loadSortingResults(self,*,key):
        obj=kb.loadObject(key=key)
        results=obj['sorting_results']
        for result in results:
            study_name=result['study_name']
            recording_name=result['recording_name']
            sorter_name=result['sorter_name']
            S=self.study(study_name)
            D=S.recording(recording_name)
            D.addSortingResult(result)
        print('Loaded {} results'.format(len(results)))
    def studyNames(self):
        return self._study_names
    def study(self,name):
        return self._studies_by_name[name]


import spiketoolkit as st
import time
import spikewidgets as sw

def mountainsort4(
        recording, # The recording extractor
        detect_sign, # Use -1, 0, or 1, depending on the sign of the spikes in the recording
        adjacency_radius, # Use -1 to include all channels in every neighborhood
        freq_min=300, # Use None for no bandpass filtering
        freq_max=6000,
        whiten=True, # Whether to do channel whitening as part of preprocessing
        clip_size=50,
        detect_threshold=3,
        detect_interval=10, # Minimum number of timepoints between events detected on the same channel
        noise_overlap_threshold=0.15, # Use None for no automated curation
        num_workers=None, # Use None for multiprocessing.cpu_count/2
        ):
    try:
        import ml_ms4alg
    except ModuleNotFoundError:
        raise ModuleNotFoundError("\nTo use Mountainsort, install ml_ms4alg: \n\n"
                                  "\npip install ml_ms4alg\n"
                                  "\nMore information on Mountainsort at: "
                                  "\nhttps://github.com/flatironinstitute/mountainsort")

    t_start_proc = time.time()

    # Bandpass filter
    if freq_min is not None:
        recording=sw.lazyfilters.bandpass_filter(recording=recording, freq_min=freq_min, freq_max=freq_max)

    # Whiten
    if whiten:
        recording=sw.lazyfilters.whiten(recording=recording)

    # Sort
    sorting=ml_ms4alg.mountainsort4(
    recording=recording,
    detect_sign=detect_sign,
    adjacency_radius=adjacency_radius,
    clip_size=clip_size,
    detect_threshold=detect_threshold,
    detect_interval=detect_interval,
    num_workers=num_workers
    )

    # Curate
    if noise_overlap_threshold is not None:
        sorting=ml_ms4alg.mountainsort4_curation(
          recording=recording,
          sorting=sorting,
          noise_overlap_threshold=noise_overlap_threshold
        )
    print('Elapsed time: ', time.time() - t_start_proc)

    return sorting



import spikeinterface as si
import spikewidgets as sw
import spiketoolkit as st
import mlprocessors as mlpr
import json
from kbucket import client as kb

def compareWithTruth(result):
    ret={}
    out=GenSortingComparisonTable.execute(firings=result['firings'],firings_true=result['firings_true'],json_out={'ext':'.json'},html_out={'ext':'.html'}).outputs
    ret['json']=kb.saveFile(out['json_out'],basename='table.json')
    ret['html']=kb.saveFile(out['html_out'],basename='table.html')
    return ret

class GenSortingComparisonTable(mlpr.Processor):
    VERSION='0.1.0'
    firings=mlpr.Input('Firings file (sorting)')
    firings_true=mlpr.Input('True firings file')
    json_out=mlpr.Output('Table as .json file produced from pandas dataframe')
    html_out=mlpr.Output('Table as .html file produced from pandas dataframe')
    
    def run(self):
        sorting=si.MdaSortingExtractor(firings_file=self.firings)
        sorting_true=si.MdaSortingExtractor(firings_file=self.firings_true)
        SC=st.comparison.SortingComparison(sorting_true,sorting)
        df=sw.SortingComparisonTable(comparison=SC).getDataframe()
        json=df.transpose().to_dict()
        html=df.to_html(index=False)
        _write_json_file(json,self.json_out)
        _write_json_file(html,self.html_out)

def _write_json_file(obj,path):
  with open(path,'w') as f:
    return json.dump(obj,f)
